
// Copyright (C) 2008 Indrek Mandre <indrek@mare.ee>
// http://www.mare.ee/indrek/irtroll/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <avr/boot.h>
#include <avr/wdt.h>

#include "xavr/xavr.h"

typedef XAVR::PortB::Out2 IROUT;
typedef XAVR::PortD::In2 BUTTON;

typedef XAVR::Pins<XAVR::PortD::In6, XAVR::PortD::In7> ACMP;

static inline bool isIROn() _XAVR_INLINE;
static inline bool isIROn()
{
  return ACSR & _BV(ACO);
}

static volatile bool overflow = false;

ISR(TIMER0_OVF_vect)
{
  overflow = true;
}

union data_str {
  struct {
    uint8_t delay;
    uint8_t cnt;
  } pair;
  uint16_t single;
};

static uint16_t freq;
#define DATALEN 127
static data_str data[DATALEN];

static EEMEM uint16_t eepromp_freq = 0;
static EEMEM data_str eeprom_data[DATALEN];

static void write_eeprom ()
{
  cli();
  eeprom_write_word (&eepromp_freq, freq);
  eeprom_write_block (&data[0], &eeprom_data[0], sizeof (data));
  sei();
}

static void read_eeprom ()
{
  cli();
  freq = eeprom_read_word (&eepromp_freq);
  eeprom_read_block (&data[0], &eeprom_data[0], sizeof (data));
  sei();
}

static void send_signal ()
{
  if ( freq == 0 )
      return;
  cli();
  OCR1A = freq;
  OCR1B = OCR1A / 2;
  TCCR1A = _BV(WGM11) | _BV(WGM10);
  TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS10);
  TIMSK1 = 0;
  sei();

  for ( uint8_t i = 0; i < DATALEN; i++ )
    {
      uint16_t delay;
      uint16_t cnt;

      if ( data[i].single & 0x8000 )
        {
          delay = data[i].single & ~0x8000;
          i++;
          cnt = data[i].single;
        }
      else
        {
          delay = data[i].pair.delay;
          cnt = data[i].pair.cnt;
        }

      // process delay
      if ( i != 0 )
        {
          TCNT0 = 0;
          while (delay)
            {
              if ( overflow )
                {
                  delay--;
                  overflow = false;
                }
            }
        }

      if ( cnt == 0 )
          return;

      while ( TCNT1 < OCR1B || TCNT1 > OCR1A - 32 ); // wait until low level
      TCCR1A |= _BV(COM1B1); // enable OC1B
      while ( TCNT1 > OCR1B ); // wait until turns on

      while (1)
        {
          while ( TCNT1 < OCR1B ); // on signal
          cnt--;
          if ( !cnt )
            {
              TCCR1A &= ~_BV(COM1B1); // disable OC1B
              break;
            }
          while ( TCNT1 > OCR1B ); // off signal
        }
    }
}

static void record_signal ()
{
  cli();
  TCCR1A = 0;
  TCCR1B = 1;
  TIMSK1 = 0;
  sei();

  uint16_t delay = 0;
  uint8_t dindex = 0;

  for ( uint8_t i = 0; i < DATALEN; i++ )
      data[i].single = 0;

  while (1)
    {
      TCNT0 = 0;
      while ( !isIROn() )
        {
          if ( overflow && delay != 0xffff )
            {
              overflow = false;
              delay++;
              if ( dindex != 0 && delay > 5000 && !BUTTON::IsOn() )
                {
                  write_eeprom ();
                  dindex = 0xff;
                  break;
                }
              else if ( dindex == 0 && !BUTTON::IsOn() ) // nothing happened
                {
                  dindex = 0xff;
                  break;
                }
            }
        }

      if ( dindex == 0xff )
          break;
      
        {
          TCNT1 = 0;
          uint16_t cnt = 0;
          uint16_t rcnt = 0;
          uint16_t period = 0;

          while (1)
            {
              TCNT0 = 0;
              overflow = false;
              while ( isIROn() && !overflow);
              if ( overflow )
                  break;

              TCNT0 = 0;
              overflow = false;
              while ( !isIROn() && !overflow );
              uint16_t t = TCNT1;
              rcnt++;
              if ( overflow )
                  break;

              if ( cnt < 100 )
                {
                  period = t;
                  cnt++;
                }
            }

          if ( cnt > 10 && dindex != DATALEN )
            {
              if ( dindex == 0 )
                {
                  freq = period / cnt;
                  delay = 0;
                }

              if ( rcnt > 0x7f || delay > 0x7f )
                {
                  if ( dindex == DATALEN - 1 )
                    {
                      dindex++;
                    }
                  else
                    {
                      if ( delay > 0x7fff )
                          delay = 0x7fff;
                      data[dindex++].single = delay | 0x8000;
                      data[dindex++].single = rcnt;
                    }
                }
              else
                {
                  data[dindex].pair.cnt = rcnt;
                  data[dindex].pair.delay = delay;
                  dindex++;
                }

              delay = 0;
            }
        }
    }
}

int main ()
{
  IROUT::Init();
  BUTTON::Init();
  ACMP::Init();

  ACSR = 0;

  TCCR0A = 0;
  TCCR0B = 1;
  TIMSK0 = _BV(TOIE0);

  sei();

  read_eeprom ();

  while (1)
    {
      uint32_t n = 1 + rand() % 40;

      for ( uint32_t i = 0; i < n * 200000; i++ )
        {
          __asm__ __volatile__ ("nop");
          if ( BUTTON::IsOn() )
              break;
        }

      if ( BUTTON::IsOn() )
        {
          // avoid bounce
          for ( uint32_t i = 0; i < 100000; i++ )
              __asm__ __volatile__ ("nop");
          if ( BUTTON::IsOn() )
              record_signal();
        }

      send_signal();
    }

  return 0;
}

